//---------------------------------------------------------------------
//  Copyright (C) Microsoft Corporation.  All rights reserved.
// 
// THIS CODE AND INFORMATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY
// KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
//---------------------------------------------------------------------

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Design;
using System.Reflection;
using System.Windows.Forms;
using System.Windows.Forms.Design;
using System.Windows.Forms.Design.Behavior;

namespace DesignSurfaceExtenderLibrary.Extensions
{
    /// <summary>
    /// Provides an abstract base implementation for the 
    /// <see cref="DesignSurfaceExtender"/> component. 
    /// </summary>
    public abstract class DesignSurfaceExtension : IDisposable
    {
        /// <summary>
        /// Backs the <see cref="ActionList"/> property.
        /// </summary>
        protected DesignerActionList actionListValue;

        private bool disposed;
        private IServiceProvider serviceProviderValue;
        private IReferenceService referenceServiceValue;
        private IComponentChangeService componentChangeServiceValue;
        private ISelectionService selectionServiceValue;
        private BehaviorService behaviorServiceValue;

        private bool enabledValue;

        /// <summary>
        /// Backs the <see cref="Adorner"/> property.
        /// </summary>
        protected Adorner adornerValue;

        /// <summary>
        /// Initializes a new instance of the <see cref="DesignSurfaceExtension"/> class.
        /// </summary>
        /// <param name="serviceProvider">An <see cref="IServiceProvider"/> which
        /// provides design-time services.</param>
        protected DesignSurfaceExtension( IServiceProvider serviceProvider )
        {
            if( serviceProvider != null )
            {
                this.serviceProviderValue = serviceProvider;
            }
            else
            {
                throw new ArgumentNullException( "serviceProvider", "must be assigned" );
            }

            this.CreateAdorner();

            // Handle the ComponentAdded and ComponentRemoved events.
            this.ComponentChangeService.ComponentAdded += 
                new ComponentEventHandler( ComponentAddedEventHandler );

            this.ComponentChangeService.ComponentRemoved += 
                new ComponentEventHandler( ComponentRemovedEventHandler );
        }

        ///////////////////////////////////////////////////////////////////////
        #region Extensibility Points

        /// <summary>
        /// Populates the <see cref="Adorner"/> with the <see cref="ExtensionGlyph"/>
        /// objects managed by this extension.
        /// </summary>
        /// <param name="control">The control to which the <see cref="ExtensionGlyph"/>
        /// objects will be bound.</param>
        /// <remarks>This method is not <c>abstract</c> because deriving types
        /// are not required to override it. For example, extensions which
        /// provide <see cref="ExtensionGlyph"/> implementations which are not
        /// bound to particular control instances may not need to implement
        /// this method.
        /// </remarks>
        protected virtual void AddGlyphsForControl( Control control )
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Removes the <see cref="ExtensionGlyph"/> objects for the given control
        /// from the related <see cref="Adorner"/>.
        /// </summary>
        /// <param name="control">The control to which the <see cref="ExtensionGlyph"/>
        /// objects will be bound.</param>
        /// <remarks>This method is not <c>abstract</c> because deriving types
        /// are not required to override it. For example, extensions which
        /// provide <see cref="ExtensionGlyph"/> implementations which are not
        /// bound to particular control instances may not need to implement
        /// this method.
        /// </remarks>
        protected virtual void RemoveGlyphsForControl( Control control )
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Creates the <see cref="Adorner"/> provided by this 
        /// extension. 
        /// </summary>
        /// <returns>A new <see cref="Adorner"/> populated with type-specific glyphs.</returns>
        private Adorner CreateAdorner()
        {
            this.adornerValue = new Adorner();
            this.adornerValue.Enabled = false;

            this.BehaviorService.Adorners.Add( this.adornerValue );

            return adornerValue;
        }

        /// <summary>
        /// Populates the <see cref="System.Windows.Forms.Design.Behavior.Adorner.Glyphs"/> collection.
        /// </summary>
        /// <param name="adorner">The <see cref="Adorner"/> to populate with glyphs.</param>
        /// <remarks>
        /// <para>
        /// Deriving classes may provide a type-specific implementation of
        /// the <see cref="PopulateAdorner"/> method.
        /// </para>
        /// <para>
        /// Implementations are expected to clear the adorner's glyph 
        /// collection before populating the collection.
        /// </para>
        /// </remarks>
        protected virtual void PopulateAdorner( Adorner adorner )
        {
            adorner.Glyphs.Clear();

            object[] controls = this.ReferenceService.GetReferences( typeof( Control ) );

            for( int i = 0; i < controls.Length; i++ )
            {
                Control c = controls[i] as Control;

                this.AddGlyphsForControl( c );
            }
        }


        /// <summary>
        /// Handles the <see cref="IComponentChangeService.ComponentAdded"/> event. 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e">A <see cref="ComponentEventArgs"/> which holds the event data.</param>
        /// <remarks>Deriving classes may provide a type-specific implementation of
        /// the <see cref="ComponentAddedEventHandler"/> method.
        /// </remarks>
        protected virtual void ComponentAddedEventHandler( object sender, ComponentEventArgs e )
        {
            if( e.Component is Control )
            {
                if( this.Enabled )
                {
                    this.PopulateAdorner( this.adornerValue );

                    // IReferenceService workaround.
                    this.AddGlyphsForControl( e.Component as Control );
                }
            }
        }

        /// <summary>
        /// Handles the <see cref="IComponentChangeService.ComponentRemoved"/> event. 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e">A <see cref="ComponentEventArgs"/> which holds the event data.</param>
        protected virtual void ComponentRemovedEventHandler( object sender, ComponentEventArgs e )
        {
            if( e.Component is Control )
            {
                this.RemoveGlyphsForControl( e.Component as Control );

                this.Invalidate();
            }
        }


        #endregion


        ///////////////////////////////////////////////////////////////////////
        #region IExtension Members

        /// <summary>
        /// Gets or sets a value indicating whether the extension is active.
        /// </summary>
        public bool Enabled
        {
            get
            {   
                return this.enabledValue;
            }

            set
            {
                // Workaround: DropBehavior enables all adorners
                // regardless of previous Enabled state, so the
                // value reported by Adorner.Enabled is unreliable.
                if( this.enabledValue != value )
                {
                    this.enabledValue = value;

                    // Workaround: DropBehavior enables all adorners
                    // regardless of previous Enabled state.
                    if( value )
                    {   
                        this.PopulateAdorner( this.adornerValue );
                    }
                    else 
                    {
                        this.adornerValue.Glyphs.Clear();
                    }

                    this.adornerValue.Enabled = this.enabledValue;
                }   
            }
        }

        /// <summary>
        /// Gets the <see cref="Adorner"/> provided by the extension. 
        /// </summary>
        //public Adorner Adorner 
        //{
        //    get
        //    {
        //        if( this.adornerValue == null )
        //        {
        //            this.adornerValue = DesignSurfaceExtension.CreateAdorner();

        //            this.PopulateAdorner( this.adornerValue );

        //            this.BehaviorService.Adorners.Add( this.adornerValue );
        //        }

        //        return this.adornerValue;
        //    }
        //}

        public void Invalidate()
        {
            this.adornerValue.Invalidate();
        }

        /// <summary>
        /// Gets the <see cref="DesignerActionList"/> provided by the extension. 
        /// </summary>
        public virtual DesignerActionList ActionList
        {
            get
            {
                return null;
            }
        }
       
        #endregion IExtension Members

        ///////////////////////////////////////////////////////////////////////
        #region IDisposable Implementation

        /// <summary>
        /// Releases all resources used by the extension. 
        /// </summary>
        /// <remarks>The <see cref="DesignSurfaceExtension"/> class implements the 
        /// the Dispose pattern in order to release event handlers.
        /// </remarks>
        public void Dispose()
        {
            Dispose( true );

            GC.SuppressFinalize( this );
        }

        /// <summary>
        /// Releases the unmanaged resources used by the extension and
        /// optionally releases the managed resources. 
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; 
        /// <c>false</c> to release only unmanaged resources. </param>
        /// <remarks>
        /// This implementation detaches the <see cref="IComponentChangeService"/>
        /// event handlers.
        /// </remarks>
        protected virtual void Dispose( bool disposing )
        {
            if( !this.disposed )
            {
                if( disposing )
                {
                    if( this.ComponentChangeService != null )
                    {
                        this.ComponentChangeService.ComponentAdded -=
                            new ComponentEventHandler( ComponentAddedEventHandler );

                        this.ComponentChangeService.ComponentRemoved -=
                            new ComponentEventHandler( ComponentRemovedEventHandler );
                    }
                }
            }

            this.disposed = true;
        }

        #endregion IDisposable Implementation

        ///////////////////////////////////////////////////////////////////////
        #region Designer Services

        /// <summary>
        /// Gets the <see cref="IServiceProvider"/> for this extension.
        /// </summary>
        protected IServiceProvider ServiceProvider
        {
            get
            {
                return this.serviceProviderValue;
            }
        }


        /// <summary>
        /// Gets a reference to the <see cref="IReferenceService"/>.
        /// </summary>
        protected IReferenceService ReferenceService
        {
            get
            {
                if( this.referenceServiceValue == null )
                {
                    this.referenceServiceValue =
                        this.ServiceProvider.GetService( typeof( IReferenceService ) )
                        as IReferenceService;
                }

                return this.referenceServiceValue;
            }
        }

        /// <summary>
        /// Gets a reference to the <see cref="IComponentChangeService"/>.
        /// </summary>
        protected IComponentChangeService ComponentChangeService
        {
            get
            {
                if( this.componentChangeServiceValue == null )
                {
                    this.componentChangeServiceValue =
                        this.ServiceProvider.GetService(typeof(IComponentChangeService))
                        as IComponentChangeService;
                }

                return componentChangeServiceValue;
            }
        }

        /// <summary>
        /// Gets a reference to the <see cref="ISelectionService"/>.
        /// </summary>
        protected ISelectionService SelectionService
        {
            get
            {
                if( this.selectionServiceValue == null )
                {
                    this.selectionServiceValue =
                        this.ServiceProvider.GetService( typeof( ISelectionService ) )
                        as ISelectionService;
                }

                return this.selectionServiceValue;
            }
        }

        /// <summary>
        /// Gets a reference to the <see cref="BehaviorService"/>.
        /// </summary>
        protected BehaviorService BehaviorService
        {
            get
            {
                if( this.behaviorServiceValue == null )
                {
                    this.behaviorServiceValue =
                        this.ServiceProvider.GetService( typeof( BehaviorService ) )
                        as BehaviorService;
                }

                return this.behaviorServiceValue;
            }
        }

        #endregion Designer Services
    }
}
